import random
import sqlite3
from contextlib import contextmanager
from pathlib import Path
from functools import partial


from ardilla import Engine, Model, Crud
from ardilla.errors import QueryExecutionError, DisconnectedEngine
from pydantic import Field

from ardilla.fields import ForeignField


path = Path(__file__).parent
db = path / "test_sync.sqlite"

unlinkdb = partial(db.unlink, missing_ok=True)

@contextmanager
def cleanup():
    unlinkdb()
    try:
        yield
    finally:
        unlinkdb()

class User(Model):
    id: int = Field(pk=True, auto=True)
    name: str

def test_context_engine():
    with cleanup():
        try:
            with Engine(db) as engine:
                crud = engine.crud(User)
                u = crud.insert(name='chris') # should pass
                assert u.name == 'chris'
            crud.insert(name='moni')
        except Exception as e:
            assert isinstance(e, DisconnectedEngine), f'Wrong exception raised'

def test_st_engine():
    unlinkdb()
    try:
        engine = Engine(db)
        engine.connect()
        crud = engine.crud(User)
        u = crud.insert(name='chris') # should pass
        assert u.name == 'chris'
        engine.close()
        crud.insert(name='moni')
    except Exception as e:
        assert isinstance(e, DisconnectedEngine), f'Wrong exception raised'
    finally:
        engine.close()
        unlinkdb()


# CREATE

def test_insert():
   with cleanup(), Engine(db) as engine:
        crud = engine.crud(User)
        u = crud.insert(name="chris")
        
        assert u is not None, "User wasn't created as expected"
        assert u.__rowid__ is not None, "Created user did not have __rowid__ set"
        assert u.__rowid__ == 1, "Created User did not have correct __rowid__ "
        try:
            crud.insert(id=1, name="chris")
        except Exception as err:
            assert isinstance(err, QueryExecutionError), f'Wrong error rised: {err}'
        else:
            raise Exception("QueryExcecutionError should have been rised")


def test_insert_or_ignore():
    with cleanup(), Engine(db) as engine:
        crud = engine.crud(User)
        kws = dict(id=1, name='chris')
        u1 = crud.insert(**kws)
        u2= crud.insert_or_ignore(**kws)
        
        assert u2 is None

def test_save_one():
    with cleanup(), Engine(db) as engine:
        crud = engine.crud(User)
        u = crud.insert(name='chris')
        u.name = 'alex'
        crud.save_one(u)
        
        user = crud.get_or_none(name='alex')
        assert user.id == 1


def test_save_many():
    users = [User(name=f'user {n}') for n in range(20)]
    with cleanup(), Engine(db) as engine:
        crud = engine.crud(User)
        crud.save_many(*users)

        assert crud.count() == 20

# READ
def test_get_all():
    with cleanup(), Engine(db) as engine:
        crud = engine.crud(User)
        for n in range(10):
            crud.insert(name=f'user {n}')

        total = crud.count()
        assert total == 10

def test_get_many():
    with cleanup(), Engine(db) as engine:
        crud = engine.crud(User)
        names = ['chris', 'moni', 'elena', 'fran']
        for name in names:
            for _ in range(3):
                crud.insert(name=name)
        
        chrises = crud.count(name='chris')
        
        assert chrises == 3

def test_get_or_create():
    with cleanup(), Engine(db) as engine:
        crud = engine.crud(User)
        chris, created = crud.get_or_create(name='chris')
        assert chris.id == 1
        assert created is True
        chris, created = crud.get_or_create(name='chris')
        assert chris.id == 1
        assert created is False

def test_get_or_none():
    with cleanup(), Engine(db) as engine:
        crud = engine.crud(User)
        chris = crud.get_or_none(name='chris')
        assert chris is None
        crud.insert(name='chris')
        chris = crud.get_or_none(name='chris')
        assert chris is not None

def test_delete_one():
    with cleanup(), Engine(db) as engine:
        crud = engine.crud(User)
        chrises = [User(name='chris') for _ in range(10)]
        crud.save_many(*chrises)
        
        x = User(id=5, name='chris')
        crud.delete_one(x)
        
        users = crud.get_all()
        assert len(users) == 9
        assert all(u.id != 5 for u in users)


def test_delete_many():
    with cleanup(), Engine(db) as engine:
        crud = engine.crud(User)
        users = [
            User(id=n, name='chris') for n in range(10)
        ]
        crud.save_many(*users)
        
        to_delete = users[:-1]
        crud.delete_many(*to_delete)
        
        users = crud.get_all()
        assert len(users) == 1, "Delete many didn't delete the correct amount of users"


def test_foreign_keys():
    db = path / 'sync_test.sqlite'
    db.unlink(missing_ok=True)
    engine = Engine(db, enable_foreing_keys=True)
    engine.connect()
    
    class Guild(Model):
        id: int = Field(pk=True, auto=True)
        name: str
        
    class User(Model):
        id: int = Field(pk=True, auto=True)
        name: str
        guild_id: int = ForeignField(references=Guild, on_delete=ForeignField.CASCADE)
    
    gcrud = engine.crud(Guild)
    ucrud = engine.crud(User)
    
    ga = gcrud.insert(name='guild a')
    gb = gcrud.insert(name='guild b')
    for guild in [ga, gb]:
        for n in range(5):
            ucrud.insert(name=f'user {n}', guild_id=guild.id)
    
    assert ucrud.count() == 10
    gcrud.delete_one(ga)
    assert ucrud.count() == 5 
    engine.close()
    db.unlink(missing_ok=True)






from __future__ import annotations
from ctypes import Union

import aiosqlite

from ..abc import BaseEngine
from ..models import M
from ..errors import DisconnectedEngine

from .crud import AsyncCrud

class AsyncEngine(BaseEngine):
    """Async Engine that uses `aiosqlite.Connection` and `aiosqlite.Cursor`
    """
    con: aiosqlite.Connection

    async def get_connection(self) -> aiosqlite.Connection:
        """Gets the connections or makes a new one but it doesn't set it as an attrib

        Returns:
            sqlite3.Connection: the connection
        """
        con: Union[aiosqlite.Connection, None] = getattr(self, 'con', None)
        if not self.check_connection():
            con: aiosqlite.Connection = await aiosqlite.connect(self.path)
            con.row_factory = aiosqlite.Row
            
            if self.enable_foreing_keys:
                await con.execute("PRAGMA foreign_keys = on;")
            
            return con
        else:
            return self.con
    
    async def connect(self) -> aiosqlite.Connection:
        """
        Stablishes a connection to the database
        Returns:
            The connection
        """
        await self.close()
        
        self.con = await self.get_connection()
        return self.con

    async def close(self) -> None:
        if self.check_connection():
            await self.con.close()
        self._cruds.clear()

    async def __aenter__(self) -> AsyncEngine:
        """Stablishes the connection and if specified enables foreign keys pragma

        Returns:
            The connection
        """
        await self.connect()
        return self

    async def __aexit__(self, *_):
        """Closes the connection"""
        await self.close()
    
    async def crud(self, Model: type[M]) -> AsyncCrud[M]:
        """
        This function works exactly like `Engine.crud` but
        returns an instance of `ardilla.asyncio.crud.AsyncCrud` instead of `ardilla.crud.Crud`
        and is asynchronous
        
        Returns:
            The async Crud for the given model
        """
        if not self.check_connection():
            raise DisconnectedEngine("Can't create crud objects with a disconnected engine")
            
        if Model.__schema__ not in self.tables_created:
            await self.con.execute(Model.__schema__)
            await self.con.commit()
            self.tables_created.add(Model.__schema__)
        
        crud = self._cruds.setdefault(Model, AsyncCrud(Model, self.con))
        return crud



"""
Contains the module's errors
"""


class ArdillaException(Exception):
    pass


class ModelIntegrityError(ArdillaException):
    pass


class MissingEngine(ArdillaException):
    pass


class QueryExecutionError(ArdillaException):
    pass


class BadQueryError(ArdillaException):
    pass


class DisconnectedEngine(ArdillaException):
    pass


disconnected_engine_error = DisconnectedEngine(
    "The engine has been disconnected and cannot operate on the database"
)


class MigrationError(ArdillaException):
    pass


from __future__ import annotations
import sqlite3
from sqlite3 import Row
from contextlib import contextmanager
from typing import Literal, Generic, Optional, Union, Generator

from . import queries
from .abc import BaseCrud
from .models import M
from .errors import BadQueryError, QueryExecutionError, disconnected_engine_error, DisconnectedEngine
from .schemas import SQLFieldType


@contextmanager
def contextcursor(con: sqlite3.Connection) -> Generator[sqlite3.Cursor, None, None]:
    """a context manager wrapper for sqlite3.Cursor

    Args:
        con (sqlite3.Connection): the connection

    Raises:
        disconnected_engine_error: if the connection is non functioning

    Yields:
        Generator[sqlite3.Cursor, None, None]: the cursor
    """
    cur = None
    try:
        cur = con.cursor()
        yield cur
    except Exception as e:
        if (
            isinstance(e, sqlite3.ProgrammingError)
            and str(e) == "Cannot operate on a closed database."
        ):
            raise DisconnectedEngine(str(e))
        else:
            raise e
    finally:
        if cur is not None:
            cur.close()


class Crud(BaseCrud, Generic[M]):
    """Abstracts CRUD actions for model associated tables"""

    connection: sqlite3.Connection

    def _do_insert(
        self,
        ignore: bool = False,
        returning: bool = True,
        /,
        **kws: SQLFieldType,
    ) -> Optional[M]:
        """private helper method for insertion methods

        Args:
            ignore (bool, optional): Ignores conflicts silently. Defaults to False.
            returning (bool, optional): Determines if the query should return the inserted row. Defaults to True.
            kws (SQLFieldType): the column name and values for the insert query

        Raises:
            QueryExecutionError: when sqlite3.IntegrityError happens because of a conflic

        Returns:
            An instance of model if any row is returned
        """
        q, vals = queries.for_do_insert(self.tablename, ignore, returning, kws)

        with contextcursor(self.connection) as cur:
            try:
                cur.execute(q, vals)
            except sqlite3.IntegrityError as e:
                raise QueryExecutionError(str(e))

            row = cur.fetchone()
            self.connection.commit()
            if returning and row:
                return self._row2obj(row, cur.lastrowid)

        return None

    def insert(self, **kws: SQLFieldType) -> M:
        """Inserts a record into the database.

        Args:
            kws (SQLFieldType): The keyword arguments are passed as the column names and values
                to the insert query

        Returns:
            Creates a new entry in the database and returns the object

        Rises:
            `ardilla.error.QueryExecutionError`: if there's a conflict when inserting the record
        """
        self.verify_kws(kws)
        return self._do_insert(False, True, **kws)

    def insert_or_ignore(self, **kws: SQLFieldType) -> Optional[M]:
        """Inserts a record to the database with the keywords passed. It ignores conflicts.

        Args:
            kws (SQLFieldType): The keyword arguments are passed as the column names and values
                to the insert query

        Returns:
            The newly created row as an instance of the model if there was no conflicts
        """
        self.verify_kws(kws)
        return self._do_insert(True, True, **kws)
    
    def get_or_none(self, **kws: SQLFieldType) -> Optional[M]:
        """Returns a row as an instance of the model if one is found or none

        Args:
            kws (SQLFieldType): The keyword arguments are passed as column names and values to
                a select query

        Example:
            ```py
            crud.get_or_none(id=42)

            # returns an object with id of 42 or None if there isn't one in the database
            ```

        Returns:
            The object found with the criteria if any
        """
        self.verify_kws(kws)
        q, vals = queries.for_get_or_none(self.tablename, kws)
        with contextcursor(self.connection) as cur:
            cur.execute(q, vals)
            row: Union[Row, None] = cur.fetchone()
            if row:
                return self._row2obj(row)
        return None

    def get_or_create(self, **kws: SQLFieldType) -> tuple[M, bool]:
        """Returns an object from the database with the spefied matching data
        Args:
            kws (SQLFieldType): the key value pairs will be used to query for an existing row
                if no record is found then a new row will be inserted
        Returns:
            A tuple with two values, the object and a boolean indicating if the
                object was newly created or not
        """
        self.verify_kws(kws)
        created = False
        result = self.get_or_none(**kws)
        if not result:
            result = self.insert_or_ignore(**kws)
            created = True
        return result, created

    def get_all(self) -> list[M]:
        """Gets all objects from the database
        Returns:
            A list with all the rows in table as instances of the model
        """
        return self.get_many()

    def get_many(
        self,
        order_by: Optional[dict[str, str]] = None,
        limit: Optional[int] = None,
        **kws: SQLFieldType,
    ) -> list[M]:
        """Queries the database and returns objects that meet the criteris

        Args:
            order_by (Optional[dict[str, str]], optional): An ordering dict. Defaults to None.
                The ordering should have the structure: `{'column_name': 'ASC' OR 'DESC'}`
                Case in values is insensitive

            limit (Optional[int], optional): The number of items to return. Defaults to None.
            kws (SQLFieldType): The column names and values for the select query

        Returns:
            a list of rows matching the criteria as intences of the model
        """
        self.verify_kws(kws)
        q, vals = queries.for_get_many(
            self.Model, 
            order_by=order_by,
            limit=limit, 
            kws=kws,
        )
        with contextcursor(self.connection) as cur:
            cur.execute(q, vals)
            rows: list[Row] = cur.fetchall()
            return [self._row2obj(row) for row in rows]

    def save_one(self, obj: M) -> Literal[True]:
        """Saves one object to the database

        Args:
            obj (M): the object to persist

        Returns:
            The literal `True` if the method ran successfuly
        """
        q, vals = queries.for_save_one(obj)
        try:
            self.connection.execute(q, vals)
            self.connection.commit()
        except:
            raise disconnected_engine_error
        return True

    def save_many(self, *objs: tuple[M]) -> Literal[True]:
        """Saves all the passed objects to the database

        Args:
            objs (M): the objects to persist

        Returns:
            The literal `True` if the method ran successfuly
        """
        q, vals = queries.for_save_many(objs)
        try:
            self.connection.executemany(q, vals)
            self.connection.commit()
        except:
            raise disconnected_engine_error

        return True

    def delete_one(self, obj: M) -> Literal[True]:
        """
        Deletes the object from the database (won't delete the actual object)
        If the object has a PK field or the rowid setup, those will be
        used to locate the obj and delete it.
        If not, this function will delete any row that meets the values of the object


        Args:
            obj (M): the object to delete

        Returns:
            The literal `True` if the method ran successfuly

        """

        q, vals = queries.for_delete_one(obj)
        try:
            self.connection.execute(q, vals)
            self.connection.commit()
        except:
            raise disconnected_engine_error
        return True

    def delete_many(self, *objs: M) -> Literal[True]:
        """
        Deletes all the objects passed

        Args:
            objs (M): the object to delete

        Returns:
            The literal `True` if the method ran successfuly

        """
        q, vals = queries.for_delete_many(objs)
        try:
            self.connection.execute(q, vals)
            self.connection.commit()
        except:
            raise disconnected_engine_error
        return True

    def count(self, column: str = '*',/, **kws) -> int:
        """Returns an integer of the number of non null values in a column
        Or the total number of rows if '*' is passed

        Args:
            column (str, optional): The column name to count rows on. 
                Defaults to '*' which counts all the rows in the table

        Returns:
            int: the number of rows with non null values in a column or the number of rows in a table
        """
        self.verify_kws(kws)
        
        tablename = self.Model.__tablename__
        if column not in self.Model.__fields__ and column != '*':
            raise BadQueryError(f'"{column}" is not a field of the "{tablename}" table')
        
        
        q, vals = queries.for_count(tablename, column, kws)
        with contextcursor(self.connection) as cur:
            cur.execute(q, vals)
            row = cur.fetchone()
    
            count = row['total_count']
                
        return count


from __future__ import annotations
from typing import Literal, Generic, Optional, Union

from typing import Any
import aiosqlite
from aiosqlite import Row

from ..errors import BadQueryError, QueryExecutionError, disconnected_engine_error
from ..models import M
from ..abc import BaseCrud
from ..schemas import SQLFieldType
from ..errors import DisconnectedEngine
from .. import queries


class ConnectionProxy:
    """A proxy class for aiosqlite.Connection that
    checks if the connections is alive before returning any of its attributes
    
    Args:
        connection (aiosqlite.Connection)
    """
    def __init__(self, connection: aiosqlite.Connection):
        self._connection = connection
    
    def __getattr__(self, __name: str) -> Any:
        if __name in {'execute', 'commit'}:
            if not self._connection._running or not self._connection._connection:
                raise DisconnectedEngine('The engine is disconnected')
        return getattr(self._connection, __name)


class AsyncCrud(BaseCrud, Generic[M]):
    """Abstracts CRUD actions for model associated tables"""
    connection: aiosqlite.Connection
    
    def __init__(self, Model: type[M], connection: aiosqlite.Connection) -> None:
        connection = ConnectionProxy(connection)
        super().__init__(Model, connection)


    async def _do_insert(
        self,
        ignore: bool = False,
        returning: bool = True,
        /,
        **kws: SQLFieldType,
    ):
        """private helper method for insertion methods

        Args:
            ignore (bool, optional): Ignores conflicts silently. Defaults to False.
            returning (bool, optional): Determines if the query should return the inserted row. Defaults to True.
            kws (SQLFieldType): the column names and values for the insertion query

        Raises:
            QueryExecutionError: when sqlite3.IntegrityError happens because of a conflic

        Returns:
            An instance of model if any row is returned
        """
        q, vals = queries.for_do_insert(self.tablename, ignore, returning, kws)

        cur = None
        try:
            cur = await self.connection.execute(q, vals)
        except aiosqlite.IntegrityError as e:
            raise QueryExecutionError(str(e))
        except aiosqlite.ProgrammingError as e:
            raise disconnected_engine_error
        else:
            row = await cur.fetchone()
            await self.connection.commit()
            if returning and row:
                return self._row2obj(row, cur.lastrowid)
        finally:
            if cur is not None:
                await cur.close()

    async def get_or_none(self, **kws: SQLFieldType) -> Optional[M]:
        """Returns a row as an instance of the model if one is found or none

        Args:
            kws (SQLFieldType): The keyword arguments are passed as column names and values to
                a select query

        Example:
            ```py
            await crud.get_or_none(id=42)

            # returns an object with id of 42 or None if there isn't one in the database
            ```
        Returns:
            The object found with the criteria if any
        """
        self.verify_kws(kws)
        q, vals = queries.for_get_or_none(self.tablename, kws)

        async with self.connection.execute(q, vals) as cur:
            row: Union[Row, None] = await cur.fetchone()
            if row:
                return self._row2obj(row)
        return None

    async def insert(self, **kws: SQLFieldType) -> M:
        """
        Inserts a record into the database.

        Args:
            kws (SQLFieldType): the column names and values for the insertion query

        Returns:
            Returns the inserted row as an instance of the model
        Rises:
            ardilla.error.QueryExecutionError: if there's a conflict when inserting the record
        """
        self.verify_kws(kws)
        return await self._do_insert(False, True, **kws)

    async def insert_or_ignore(self, **kws: SQLFieldType) -> Optional[M]:
        """Inserts a record to the database with the keywords passed. It ignores conflicts.

        Args:
            kws (SQLFieldType): The keyword arguments are passed as the column names and values
                to the insert query

        Returns:
            The newly created row as an instance of the model if there was no conflicts
        """
        self.verify_kws(kws)
        return await self._do_insert(True, True, **kws)

    async def get_or_create(self, **kws: SQLFieldType) -> tuple[M, bool]:
        """Returns an object from the database with the spefied matching data
        Args:
            kws (SQLFieldType): the key value pairs will be used to query for an existing row
                if no record is found then a new row will be inserted
        Returns:
            A tuple with two values, the object and a boolean indicating if the
                object was newly created or not
        """
        created = False
        result = await self.get_or_none(**kws)
        if not result:
            result = await self.insert_or_ignore(**kws)
            created = True
        return result, created

    async def get_all(self) -> list[M]:
        """Gets all objects from the database

        Returns:
            A list with all the rows in table as instances of the model
        """
        q = f"SELECT rowid, * FROM {self.tablename};"

        async with self.connection.execute(q) as cur:
            return [self._row2obj(row) for row in await cur.fetchall()]

    async def get_many(
        self,
        order_by: Optional[dict[str, str]] = None,
        limit: Optional[int] = None,
        **kws: SQLFieldType,
    ) -> list[M]:
        """Queries the database and returns objects that meet the criteris

        Args:
            order_by (Optional[dict[str, str]], optional): An ordering dict. Defaults to None.
                The ordering should have the structure: `{'column_name': 'ASC' OR 'DESC'}`
                Case in values is insensitive
            kws (SQLFieldType): the column names and values for the select query

            limit (Optional[int], optional): The number of items to return. Defaults to None.

        Returns:
            a list of rows matching the criteria as intences of the model
        """
        self.verify_kws(kws)

        q, vals = queries.for_get_many(
            self.Model, order_by=order_by, limit=limit, kws=kws
        )
        async with self.connection.execute(q, vals) as cur:
            rows: list[Row] = await cur.fetchall()
            return [self._row2obj(row) for row in rows]

    async def save_one(self, obj: M) -> Literal[True]:
        """Saves one object to the database

        Args:
            obj (M): the object to persist

        Returns:
            The literal `True` if the method ran successfuly
        """
        q, vals = queries.for_save_one(obj)

        await self.connection.execute(q, vals)
        await self.connection.commit()
        return True

    async def save_many(self, *objs: M) -> Literal[True]:
        """Saves all the passed objects to the database

        Args:
            objs (M): the objects to persist

        Returns:
            The literal `True` if the method ran successfuly
        """
        q, vals = queries.for_save_many(objs)

        await self.connection.executemany(q, vals)
        await self.connection.commit()

        return True

    async def delete_one(self, obj: M) -> Literal[True]:
        """
        Deletes the object from the database (won't delete the actual object)
        If the object has a PK field or the rowid setup, those will be
        used to locate the obj and delete it.
        If not, this function will delete any row that meets the values of the object


        Args:
            obj (M): the object to delete

        Returns:
            The literal `True` if the method ran successfuly

        """
        q, vals = queries.for_delete_one(obj)

        await self.connection.execute(q, vals)
        await self.connection.commit()
        return True

    async def delete_many(self, *objs: M) -> Literal[True]:
        """
        Deletes all the objects passed

        Args:
            objs (M): the object to delete

        Returns:
            The literal `True` if the method ran successfuly

        """
        q, vals = queries.for_delete_many(objs)

        await self.connection.execute(q, vals)
        await self.connection.commit()
        
    async def count(self, column: str = '*', /, **kws) -> int:
        """Returns an integer of the number of non null values in a column
        Or the total number of rows if '*' is passed

        Args:
            column (str, optional): The column name to count rows on. 
                Defaults to '*' which counts all the rows in the table

        Returns:
            int: the number of rows with non null values in a column or the number of rows in a table
        """
        tablename = self.Model.__tablename__
        if column not in self.Model.__fields__ and column != '*':
            raise BadQueryError(f'"{column}" is not a field of the "{tablename}" table')
        
        q, vals = queries.for_count(tablename, column, kws)
        async with self.connection.execute(q, vals) as cur:
            row = await cur.fetchone()    
            count = row['total_count']
                
        return count


from __future__ import annotations
import sqlite3
from typing import Any, Literal, TypeVar, Optional, Union
from abc import abstractmethod, ABC
from sqlite3 import Row

from .models import M, Model as BaseModel

E = TypeVar("E")  # Engine Type

Connection = TypeVar("Connection")
CrudType = TypeVar('CrudType', bound='BaseCrud')


class BaseEngine(ABC):
    """This just provides autocompletition across the library"""

    __slots__ = (
        "path",  # the path to the database
        "schemas",  # the registered tables
        "tables_created",  # a list of tables that were setup
        "enable_foreing_keys",  # a bool to specify if the pragma should be enforced
        "con", # sync connection
        "_cruds", # crud cache
    )
    
    def check_connection(self) -> bool:
        """Checks if the engine's connection is alive
        works for both the sync and async classes

        Returns:
            bool: if the connection is fine
        """
        con: Union[Connection, None] = getattr(self, 'con', None)
        try:
            if isinstance(con, sqlite3.Connection):
                con.cursor()
                return True
            elif con is not None:
                # should be aiosqlite
                # we don't import it here to prevent import errors 
                # in case there's missing dependency of aiosqlite
                return con._running and con._connection
            else:
                return None
        except:
            return False

    def __init__(
        self,
        path: str,
        enable_foreing_keys: bool = False,
    ):
        self.path = path
        self.schemas: set[str] = set()
        self.tables_created: set[str] = set()
        self._cruds: dict[type[M], CrudType] = {}
        self.enable_foreing_keys = enable_foreing_keys
    
    @abstractmethod
    def get_connection(self) -> Connection:
        ...
        
    @abstractmethod
    def connect(self) -> Connection:
        ...
        
    @abstractmethod
    def close(self) -> None:
        ...

    @abstractmethod
    def crud(self, Model: type[M]) -> CrudType:
        ...


class BaseCrud(ABC):
    __slots__ = (
        "connection",
        "tablename",
        "Model",
        "columns",
    )

    def __init__(self, Model: type[M], connection: Connection) -> None:
        self.Model = Model
        self.connection = connection

        self.tablename = Model.__tablename__
        self.columns = tuple(Model.__fields__)

    def __new__(cls, Model: type[M], connection: Connection):
        if not issubclass(Model, BaseModel):
            raise TypeError("Model param has to be a subclass of model")

        return super().__new__(cls)

    def verify_kws(self, kws: dict[str, Any]) -> Literal[True]:
        """Verifies that the passed kws keys in dictionary
        are all contained within the model's fields

        Args:
            kws (dict[str, Any]): the keyword arguments for queries

        Returns:
            Literal[True]: If the kws are verified
        """
        for key in kws:
            if key not in self.Model.__fields__:
                raise KeyError(
                    f'"{key}" is not a field of the "{self.Model.__name__}" and cannot be used in queries'
                )
        return True

    def _row2obj(self, row: Row, rowid: Optional[int] = None) -> BaseModel:
        """
        Args:
            row: the sqlite row
            rowid: the rowid of the row.
                If passed it means it comes from an insert function

        """
        keys = list(self.Model.__fields__)
        if rowid is None:
            rowid, *vals = row
        else:
            vals = list(row)
        data = {k: v for k, v in zip(keys, vals)}

        obj = self.Model(**data)
        obj.__rowid__ = rowid
        return obj

    # Create
    @abstractmethod
    def _do_insert(self, ignore: bool = False, returning: bool = True, /, **kws):
        ...

    @abstractmethod
    def insert(self, **kws):
        ...

    @abstractmethod
    def insert_or_ignore(self):
        ...

    # Read
    @abstractmethod
    def get_all(self) -> list[M]:
        ...

    @abstractmethod
    def get_many(
        self,
        order_by: Optional[dict[str, str]] = None,
        limit: Optional[int] = None,
        **kws,
    ) -> list[M]:
        ...

    @abstractmethod
    def get_or_create(self, **kws) -> tuple[M, bool]:
        ...

    @abstractmethod
    def get_or_none(self, **kws) -> Optional[M]:
        ...

    # Update
    @abstractmethod
    def save_one(self, obj: M) -> Literal[True]:
        ...

    @abstractmethod
    def save_many(self, *objs: M) -> Literal[True]:
        ...

    # Delete
    @abstractmethod
    def delete_one(self, obj: M) -> Literal[True]:
        ...

    @abstractmethod
    def delete_many(self, *objs: M) -> Literal[True]:
        ...

    @abstractmethod
    def count(self, column: str = '*', /, **kws) -> int:
        ...


from __future__ import annotations
import sqlite3
from typing import Union

from .models import M
from .crud import Crud
from .abc import BaseEngine
from .errors import DisconnectedEngine

class Engine(BaseEngine):
    """The sync engine that uses `sqlite3.Connection` and `sqlite3.Cursor`
    
    Args:
        path (str): a pathlike object that points to the sqlite database
        enable_foreing_keys (bool, optional): specifies if the pragma should be enforced. Defaults to False.
    
    Attributes:
        path (str): the path to the db
        schemas (set[str]): a set of table schemas
        tables_created (set[str]): the tables that have been setup by the engine
        enable_foreing_keys (bool): if True, the engine enables the pragma on all connections
    """
    con: sqlite3.Connection
    
    def __init__(self, path: str, enable_foreing_keys: bool = False):
        super().__init__(path, enable_foreing_keys)
    
    def get_connection(self) -> sqlite3.Connection:
        """Gets the connections or makes a new one but it doesn't set it as an attrib

        Returns:
            sqlite3.Connection: the connection
        """
        con: Union[sqlite3.Connection, None] = getattr(self, 'con', None)
        if not self.check_connection():
            con = sqlite3.connect(self.path)
            con.row_factory = sqlite3.Row
            
            if self.enable_foreing_keys:
                con.execute("PRAGMA foreign_keys = on;")
            
            return con
        else:
            return self.con
        
    def __enter__(self):
        self.connect()
        return self
    
    def __exit__(self, *_):
        self.close()
        
    def connect(self) -> sqlite3.Connection:
        self.close()
        self.con = self.get_connection()
        return self.con

    def close(self) -> None:
        if self.check_connection():
            self.con.close()
        self._cruds.clear()
    
    def crud(self, Model: type[M]) -> Crud[M]:
        """returns a Crud instances for the given model type

        Args:
            Model (type[M]): the model type for the crud object

        Returns:
            Crud[M]: the crud for the model type
        """
        if not self.check_connection():
            raise DisconnectedEngine("Can't create crud objects with a disconnected engine")
        
        if Model.__schema__ not in self.tables_created:
            self.con.execute(Model.__schema__)
            self.con.commit()
            self.tables_created.add(Model.__schema__)
        
        crud = self._cruds.setdefault(Model, Crud(Model, self.con))
        
        return crud




from typing import Any
from pydantic import Field
from ardilla import Model

class _ForeignFieldMaker():
    """
    Helper class to generate foreing key field constrains.
    
    Intead of instantiating this class the developer should use 
    the already instantiated `ardilla.fields.ForeignKey`
    instead of directly instantiating this class.
    
    Attributes:
        NO_ACTION (str): (class attribute) The database won't take action. This most likely will result in errors
        RESTRICT (str): (class attribute) The app will not be able to delete the foreing row unless there's no related child elements left
        SET_NULL (str): (class attribute) The app will set the child to Null if the parent is deleted
        SET_DEFAULT (str): (class attribute) Returns the value of this field to the default of the child when the parent is deleted or updated
        CASCADE (str): (class attribute) If the parent gets deleted or updated the child follows  
        
    """
    NO_ACTION = 'NO ACTION'
    RESTRICT = 'RESTRICT'
    SET_NULL = 'SET NULL'
    SET_DEFAULT = 'SET DEFAULT'
    CASCADE = 'CASCADE'
    
    def __call__(
        self,
        *,
        references: type[Model],
        on_delete: str = NO_ACTION, 
        on_update: str = NO_ACTION,
        **kws,
    ) -> Any:
        """
        Args:
            references (type[Model]):
                The model this foreign key points to
            on_delete (str): defaults to 'NO ACTION'
                what happens when the referenced row gets deleted
            on_update (str): defaults to 'NO ACTION'
                what happens when the referenced row gets updated
        Returns:
            A `pydantic.Field` with extra metadata for the schema creation
        Raises:
            KeyError: if the referenced value is not a type of model
            ValueError: if the referenced model does not have a primary key or has not yet been instantiated
        """
        if not issubclass(references, Model):
            raise TypeError('The referenced type must be a subclass of ardilla.Model')
        fk = getattr(references, '__pk__', None)
        tablename = getattr(references, '__tablename__')
        
        if not fk:
            raise ValueError('The referenced model requires to have a primary key')
        
        return Field(
            references=tablename, 
            fk=fk,
            on_delete=on_delete,
            on_update=on_update,
            **kws
        )

ForeignField = _ForeignFieldMaker()


"""
variables and functions here are used to generate and work with the Model's schemas
"""
import re
from typing import Optional, Union
from datetime import datetime, date, time
from pydantic import BaseModel, Json
from pydantic.fields import ModelField

from .errors import ModelIntegrityError


SCHEMA_TEMPLATE: str = "CREATE TABLE IF NOT EXISTS {tablename} (\n{fields}\n);"

SQLFieldType = Union[int, float, str, bool, datetime, bytes, date, time]

FIELD_MAPPING: dict[type, str] = {
    int: "INTEGER",
    float: "REAL",
    str: "TEXT",
    bool: "INTEGER",
    datetime: "DATETIME",
    bytes: "BLOB",
    date: "DATE",
    time: "TIME",
}

AUTOFIELDS = {
    int: " AUTOINCREMENT",
    datetime: " DEFAULT CURRENT_TIMESTAMP",
    date: " DEFAULT CURRENT_DATE",
    time: " DEFAULT CURRENT_TIME",
}


def get_tablename(model: type[BaseModel]) -> str:
    """returns the tablename of a model either from the attribute __tablenam__
    or from the lowercase model's name

    Args:
        model (type[BaseModel]): the model

    Returns:
        str: the name of the table
    """
    return getattr(model, "__tablename__", model.__name__.lower())


def make_field_schema(field: ModelField) -> dict:
    output = {}
    name = field.name
    T = field.type_
    default = field.default
    extra = field.field_info.extra
    auto = output["auto"] = extra.get("auto")
    unique = output["unique"] = extra.get("unique")
    is_pk = False
    constraint = None

    if default and unique:
        raise ModelIntegrityError(
            "field {name} has both unique and default constrains which are incompatible"
        )

    autoerror = ModelIntegrityError(
        f'field {name} has a type of "{T}" which does not support "auto"'
    )
    schema = f"{name} {FIELD_MAPPING[T]}"

    primary_field_keys = {"pk", "primary", "primary_key"}
    if len(extra.keys() & primary_field_keys) > 1:
        raise ModelIntegrityError(f'Multiple keywords for a primary field in "{name}"')

    for k in primary_field_keys:
        if k in extra and extra[k]:
            is_pk = True

            schema += " PRIMARY KEY"

            if auto and T in AUTOFIELDS:
                schema += AUTOFIELDS[T]
                field.required = (
                    False  # to allow users to create the objs without this field
                )

            elif auto:
                raise autoerror

            break
    else:
        if auto and T in AUTOFIELDS.keys() - {int}:
            schema += AUTOFIELDS[T]
        elif auto:
            raise autoerror
        elif default is not None:
            if T in {int, str, float, bool}:
                schema += f" DEFAULT {default!r}"
            elif T in {datetime, date, time}:
                schema += f" DEFAULT {default}"
            elif T is bytes:
                schema += f" DEFAULT (X'{default.hex()}')"
        elif field.required:
            schema += " NOT NULL"
        if unique:
            schema += " UNIQUE"

    if extra.get("references"):
        references, fk, on_delete, on_update = (
            extra.get(f) for f in ["references", "fk", "on_delete", "on_update"]
        )
        constraint = (
            f"FOREIGN KEY ({name}) "
            f"REFERENCES {references}({fk}) "
            f"ON UPDATE {on_update} "
            f"ON DELETE {on_delete}"
        )

    output.update({"pk": is_pk, "schema": schema, "constraint": constraint})

    return output


def make_table_schema(Model: type[BaseModel]) -> str:
    tablename = get_tablename(Model)
    fields = []
    constrains = []
    pk = None
    for field in Model.__fields__.values():
        name = field
        field_schema = make_field_schema(field)
        if field_schema["pk"] is True:
            if pk is not None:
                raise ModelIntegrityError(
                    f'field "{name}" is marked as primary but there is already a primary key field "{pk}"'
                )
            pk = field.name
        fields.append(field_schema["schema"])

        constrains.append(field_schema["constraint"]) if field_schema[
            "constraint"
        ] else None

    schema = (
        f"CREATE TABLE IF NOT EXISTS {tablename}(\n"
        + ",\n".join(f"\r    {f}" for f in (fields + constrains))
        + "\n);"
    )
    return schema


def get_pk(schema: str) -> Optional[str]:
    """Gets the primary key field name from the passed schema

    Args:
        schema (str): table schema

    Returns:
        Optional[str]: the name of the primary key if any
    """
    # Check if the schema contains a primary key definition
    if "PRIMARY KEY" in schema:
        # Use a regular expression to extract the primary key column name
        match = re.search(r"(?i)\b(\w+)\b\s+(?:\w+\s+)*PRIMARY\s+KEY", schema)
        if match:
            return match.group(1)
    return None



from typing import Optional


from .models import Model
from .errors import MigrationError
from .schemas import make_field_schema, make_table_schema



def generate_migration_script(
    old: type[Model],
    new: type[Model],
    *,
    original_tablename: str,
    new_tablename: Optional[str] = None
) -> str:
    """_summary_

    Args:
        old (type[Model]): the old model
        new (type[Model]): the new model
        original_tablename (str): the tablename as it is in the database before migrating
        new_tablename (Optional[str], optional): If the table should change its name this is the new one. Defaults to None.

    Raises:
        MigrationError: Migration includes a new field with unique constraint
        MigrationError: Migration includes a new field with primary key constraint
        MigrationError: Migration includes a not null field without a default value

    Returns:
        str: The migration script. Execute it with an sqlite3 connection
    """      
    scripts = []

    if new_tablename is not None:
        scripts.append(
            f"ALTER TABLE {original_tablename} RENAME TO {new_tablename};"
        )
    
    tablename = tablename if not new_tablename else new_tablename

    old_fields = set(old.__fields__)
    new_fields = set(new.__fields__)

    dropped = old_fields - new_fields
    for field_name in dropped:
        scripts.append(f"ALTER TABLE {tablename} DROP COLUMN {field_name};")

    added = new_fields - old_fields
    for field_name in added:
        field = new.__fields__[field_name]
        schema = make_field_schema(field)
        if schema["unique"]:
            raise MigrationError(
                f"cannot process '{field_name}' because it's marked as unique"
            )
            continue
        if schema["pk"]:
            raise MigrationError(
                f"cannot process '{field_name}' because it's marked as primary key"
            )
        field_schema = schema["schema"]
        if "NOT NULL" in field_schema and not "DEFAULT" in field_schema:
            raise MigrationError(
                f'Cannot script a "not null" field without default value in field "{field_name}"'
            )

        scripts.append(f"ALTER TABLE {tablename} ADD COLUMN {field_schema};")

    conserved = old_fields & new_fields
    alter_fields = False
    for f in conserved:
        old_schema = make_field_schema(old.__fields__[f])
        new_schema = make_field_schema(new.__fields__[f])
        if old_schema != new_schema:
            alter_fields = True
            
            # if old.__fields__[f].type_ != new.__fields__[f].type_:
            #     print(
            #         f"Ardilla can't handle type changes for now. "
            #         f"You'll have to migrate this on your own."
            #     )
            #     alter_fields = False
            #     break

    if alter_fields is True:
        new_table_schema = make_table_schema(new)
        cols = ', '.join(name for name in new.__fields__)

        script = f'''
        \rALTER TABLE {tablename} RENAME TO _{tablename};
        \r
        \r{new_table_schema}
        \r
        \rINSERT INTO {tablename} ({cols})
        \r  SELECT {cols}
        \r  FROM _{tablename};
        \r
        \rDROP TABLE _{tablename};
        \r'''

        scripts.append(script)
        

    return "\n\n".join(scripts)





from contextlib import asynccontextmanager
from pathlib import Path
from functools import partial

import pytest

from ardilla import Model, Field, ForeignField
from ardilla.asyncio import Engine
from ardilla.errors import QueryExecutionError, DisconnectedEngine




path = Path(__file__).parent
db = path / "test_sync.sqlite"

unlinkdb = partial(db.unlink, missing_ok=True)

@asynccontextmanager
async def cleanup():
    unlinkdb()
    try:
        yield
    finally:
        unlinkdb()

class User(Model):
    id: int = Field(pk=True, auto=True)
    name: str


@pytest.mark.asyncio
async def test_context_engine():
    async with cleanup():
        try:
            async with Engine(db) as engine:
                crud = await engine.crud(User)
                u = await crud.insert(name='chris') # should pass
                assert u.name == 'chris'
            await crud.insert(name='moni')
        except Exception as e:
            assert isinstance(e, DisconnectedEngine), f'Wrong exception raised'

@pytest.mark.asyncio
async def test_st_engine():
    unlinkdb()
    try:
        engine = Engine(db)
        await engine.connect()
        crud = await engine.crud(User)
        u = await crud.insert(name='chris') # should pass
        assert u.name == 'chris'
        await engine.close()
        await crud.insert(name='moni')
    except Exception as e:
        assert isinstance(e, DisconnectedEngine), f'Wrong exception raised'
    finally:
        await engine.close()
        unlinkdb()


# CREATE

@pytest.mark.asyncio
async def test_insert():
    async with cleanup(), Engine(db) as engine:
        crud = await engine.crud(User)
        u = await crud.insert(name="chris")
        
        assert u is not None, "User wasn't created as expected"
        assert u.__rowid__ is not None, "Created user did not have __rowid__ set"
        assert u.__rowid__ == 1, "Created User did not have correct __rowid__ "
        try:
            await crud.insert(id=1, name="chris")
        except Exception as err:
            assert isinstance(err, QueryExecutionError), f'Wrong error rised: {err}'
        else:
            raise Exception("QueryExcecutionError should have been rised")

@pytest.mark.asyncio
async def test_insert_or_ignore():
    async with cleanup(), Engine(db) as engine:
        crud = await engine.crud(User)
        kws = dict(id=1, name='chris')
        await crud.insert(**kws)
        u2 = await crud.insert_or_ignore(**kws)
        
        assert u2 is None


@pytest.mark.asyncio
async def test_save_one():
    async with cleanup(), Engine(db) as engine:
        crud = await engine.crud(User)
        u = await crud.insert(name='chris')
        u.name = 'alex'
        await crud.save_one(u)
        
        user = await crud.get_or_none(name='alex')
        assert user.id == 1

@pytest.mark.asyncio
async def test_save_many():
    users = [User(name=f'user {n}') for n in range(20)]
    async with cleanup(), Engine(db) as engine:
        crud = await engine.crud(User)
        await crud.save_many(*users)

        assert await crud.count() == 20

# READ
@pytest.mark.asyncio
async def test_get_all():
    async with cleanup(), Engine(db) as engine:
        crud = await engine.crud(User)
        for n in range(10):
            await crud.insert(name=f'user {n}')

        assert await crud.count() == 10

@pytest.mark.asyncio
async def test_get_many():
    async with cleanup(), Engine(db) as engine:
        crud = await engine.crud(User)
        names = ['chris', 'moni', 'elena', 'fran']
        for name in names:
            for _ in range(3):
                await crud.insert(name=name)
                
        assert await crud.count(name='chris') == 3

@pytest.mark.asyncio
async def test_get_or_create():
    async with cleanup(), Engine(db) as engine:
        crud = await engine.crud(User)
        chris, created = await crud.get_or_create(name='chris')
        assert chris.id == 1
        assert created is True
        chris, created = await crud.get_or_create(name='chris')
        assert chris.id == 1
        assert created is False

@pytest.mark.asyncio
async def test_get_or_none():
    async with cleanup(), Engine(db) as engine:
        crud = await engine.crud(User)
        chris = await crud.get_or_none(name='chris')
        assert chris is None
        await crud.insert(name='chris')
        chris = await crud.get_or_none(name='chris')
        assert chris is not None

@pytest.mark.asyncio
async def test_delete_one():
    async with cleanup(), Engine(db) as engine:
        crud = await engine.crud(User)
        chrises = [User(name='chris') for _ in range(10)]
        await crud.save_many(*chrises)
        
        x = User(id=5, name='chris')
        await crud.delete_one(x)
        
        users = await crud.get_all()
        assert len(users) == 9
        assert all(u.id != 5 for u in users)

@pytest.mark.asyncio
async def test_delete_many():
    async with cleanup(), Engine(db) as engine:
        crud = await engine.crud(User)
        users = [
            User(id=n, name='chris') for n in range(10)
        ]
        await crud.save_many(*users)
        
        to_delete = users[:-1]
        await crud.delete_many(*to_delete)
        
        assert await crud.count() == 1, "Delete many didn't delete the correct amount of users"

@pytest.mark.asyncio
async def test_foreign_keys():
    db = path / 'sync_test.sqlite'
    db.unlink(missing_ok=True)
    engine = Engine(db, enable_foreing_keys=True)
    await engine.connect()
    
    class Guild(Model):
        id: int = Field(pk=True, auto=True)
        name: str
        
    class User(Model):
        id: int = Field(pk=True, auto=True)
        name: str
        guild_id: int = ForeignField(references=Guild, on_delete=ForeignField.CASCADE)
    
    gcrud = await engine.crud(Guild)
    ucrud = await engine.crud(User)
    
    ga = await gcrud.insert(name='guild a')
    gb = await gcrud.insert(name='guild b')
    for guild in [ga, gb]:
        for n in range(5):
            await ucrud.insert(name=f'user {n}', guild_id=guild.id)
    
    assert await ucrud.count() == 10
    await gcrud.delete_one(ga)
    assert await ucrud.count() == 5 
    await engine.close()
    db.unlink(missing_ok=True)

